// VERSION 1.0ss
`timescale 1ns / 1ps

`define MEMORY_IMAGE_FILE "breakout.hex"

/* A memory with two read ports and two read-write ports (plus the video memory
 * read port), for use in a superscalar pipeline. The memory accomplishes this
 * by using the interaction between the clock and the global write-enable
 * signal. 'i' suffixes are for instruction memory, 'd' are for data memory,
 * and 'v' are for video memory; 1 and 2 suffixes denote ports 1 and 2.
 *
 * The behavior of the dual memory ports is that the operations for port 1
 * always happen before the operations for port 2. For example, if there is a
 * data load on port 1 from the same address as a data store on port 2, then
 * the load receives the old value from memory. Conversely, if there is a data
 * store on port 1 to the same address as a data load on port 2, then the load
 * receives the new value that was just stored.
 * 
 * When the instruction memory reads from an address that is being stored to
 * in the same cycle (both port 1 or both port 2), the instruction memory gets
 * the old value. This case doesn't really matter though, because memory
 * locations holding instructions should never be stored to anyway.
 */
module bram_ss(
    rst, gwe,
    clk_id, clk_v,
    addr_i1, addr_i2, addr_d1, addr_d2, addr_v,
    we_d1, we_d2,
    din_d1, din_d2,
    dout_i1, dout_i2, dout_d1, dout_d2, dout_v);
    
    input           rst;
    input           gwe;
    input           clk_id;
    input           clk_v;
    input [15:0]    addr_i1;
    input [15:0]    addr_i2;
    input [15:0]    addr_d1;
    input [15:0]    addr_d2;
    input [15:0]    addr_v;
    input           we_d1;
    input           we_d2;
    input [15:0]    din_d1;
    input [15:0]    din_d2;
    output [15:0]   dout_i1;
    output [15:0]   dout_i2;
    output [15:0]   dout_d1;
    output [15:0]   dout_d2;
    output [15:0]   dout_v;
    
    wire [15:0]     dout_v;
    wire            pipe_gwe_1, pipe_gwe_2, pipe_gwe_3;
    wire [15:0]     pipe_mem_out_i, pipe_mem_out_d;
    wire [15:0]     mem_addr_i;
    wire [15:0]     mem_addr_d;
    wire            mem_we_d;
    wire [15:0]     mem_din_d;
    
    reg [15:0]      RAM [65535:0];
    reg [15:0]      mem_out_i;
    reg [15:0]      mem_out_d;
    reg [15:0]      mem_out_v;

    integer         f;
    initial begin
        f = 0;
        // check that memory image exists
        f = $fopen( `MEMORY_IMAGE_FILE, "r" );
        if (f == 0) begin
            $display( "Memory image file: %s not found.", `MEMORY_IMAGE_FILE );
            $stop;
            end
        $fclose( f );
        // load memory image
        $readmemh(`MEMORY_IMAGE_FILE, RAM, 0, 65535);
        end
	  
    //For the "double-pumping" logic, we want to use a "gwe" signal that's high
    //in the FIRST (not the last) cycle of the four-cycle "segment" before the
    //registers are written.
    wire         new_gwe;
    //We need the new_GWE signal to reset to high/enabled so that it is active
    //on the very first clock cycle that the global reset signal goes low.
    one_pulse_rst_high new_GWE_pulser (
        .clk(clk_id),
        .rst(rst),
        .btn(gwe),
        .pulse_out(new_gwe));
    /* We "pipeline" the new global write-enable signal to give us enough state
     * information to get the memory timing right. The gwe signal is high for
     * one cycle out of every four clock cycles. In the first cycle of a
     * four-cycle "section," the new gwe is high and the address inputs to the
     * memory change. In the second cycle, the port 1 operations are performed;
     * however, the results of these operations are not immediately ouput, but
     * are pipelined for one cycle. In the third cycle, the port 2 operations
     * are performed, and the results for both the port 1 and port 2 operations
     * are now output. Nothing happens in the fourth cycle, and the outputs
     * remain stable.
     */
    Nbit_reg #(1,0) reg_gwe_1(
        .in(new_gwe),
        .out(pipe_gwe_1),
        .clk(clk_id),
        .we(1'b1),
        .gwe(1'b1),
        .rst(1'b0));
    Nbit_reg #(1,0) reg_gwe_2(
        .in(pipe_gwe_1),
        .out(pipe_gwe_2),
        .clk(clk_id),
        .we(1'b1),
        .gwe(1'b1),
        .rst(1'b0));
    Nbit_reg #(1,0) reg_gwe_3(
        .in(pipe_gwe_2),
        .out(pipe_gwe_3),
        .clk(clk_id),
        .we(1'b1),
        .gwe(1'b1),
        .rst(1'b0));
    
    /* Use the pipelined gwe signal to determine which values, for port "1" or
     * port "2," to use for the memory. There are not really two sets of ports
     * on the memory; we just use the pipelined gwe signal to mux in the proper
     * inputs in two consecutive cycles, effectively "double-pumping" the
     * memory during the four-cycle gwe "section." The values below are only
     * used in the second and third cycles of the four-cycle gwe section, so
     * use the pipelined gwe signal to enable the memory when we actually want
     * to read from it (in the always block below).
     */
    assign mem_addr_i   = (pipe_gwe_1) ? addr_i1 : addr_i2;
    assign mem_addr_d   = (pipe_gwe_1) ? addr_d1 : addr_d2;
    assign mem_we_d     = (pipe_gwe_1) ? we_d1   : we_d2;
    assign mem_din_d    = (pipe_gwe_1) ? din_d1  : din_d2;
    
    //Behavioral code for the actual memory:
    //Instruction/data memory:
    always @(posedge clk_id) begin
        //Only enable memory for the two cycles we want to use:
        if (pipe_gwe_1 || pipe_gwe_2) begin
            if (mem_we_d)
                RAM[mem_addr_d] <= mem_din_d;
            mem_out_i   <= RAM[mem_addr_i];
            mem_out_d   <= RAM[mem_addr_d];
            end
        end
    //Video memory: always enabled.
    always @(posedge clk_v) begin
        mem_out_v   <= RAM[addr_v];
        end

    //For instruction port 1 and data port 1, output (and latch) the value
    //read from memory at the end of the cycle that it was read.
    Nbit_reg #(16,0) reg_dout_i1(
        .in(mem_out_i),
        .out(dout_i1),
        .clk(clk_id),
        .we(pipe_gwe_2),
        .gwe(1'b1),
        .rst(1'b0));
    Nbit_reg #(16,0) reg_dout_d1(
        .in(mem_out_d),
        .out(dout_d1),
        .clk(clk_id),
        .we(pipe_gwe_2),
        .gwe(1'b1),
        .rst(1'b0));
    //For instruction port 2 and data port 2, output the value read from memory
    //as soon as it is read; this is the same time that the values for port 1
    //are output also. At the end of the cycle, latch the values read from
    //memory.
    Nbit_reg #(16,0) reg_mem_out_i(
        .in(mem_out_i),
        .out(pipe_mem_out_i),
        .clk(clk_id),
        .we(pipe_gwe_3),
        .gwe(1'b1),
        .rst(1'b0));
    Nbit_reg #(16,0) reg_mem_out_d(
        .in(mem_out_d),
        .out(pipe_mem_out_d),
        .clk(clk_id),
        .we(pipe_gwe_3),
        .gwe(1'b1),
        .rst(1'b0));
    assign dout_i2  = (pipe_gwe_3) ? mem_out_i : pipe_mem_out_i;
    assign dout_d2  = (pipe_gwe_3) ? mem_out_d : pipe_mem_out_d;
    assign dout_v   = mem_out_v;
    
endmodule // bram_ss
