`default_nettype none

// VERSION 1.1
  
/* This module is a ms-granularity timer device. This module has 2
   "virtual register": the timer interval register (TIR) and the timer
   status register (TSR). The TIR specifies the interval, in
   *MILLISECONDS*, between timer "ticks." The TSR has the value x8000
   if a tick has occurred since it was last read; reading the TSR
   resets it to x0000 so that it can display the next tick. The TSR
   doesn't tell you how many ticks have happened since you last read
   it, only whether a tick occurred.

   The `selectTM[RI]' inputs choose the TSR and the TIR. `selectInterval'
   and `selectTSR' should never both be high at the same time; this
   will yield garbage output. They can both be low at the same time,
   however. The TSR can only be read; the TIR can be read and written:
   which operation is performed on the TIR is governed by the `we'
   signal.

   The `in' signal is the value to be written to the TIR (it's ignored
   if `we' is low). The `out' signal is the output of the register
   indicated by `select'. Note that if you write to the TIR, the new
   value will appear on `out'.
 
   Internally, a simple counter (register + incrementer) keeps track
   of elapsed clock periods, and this value is compared with the TIR
   to see if a "tick" has occurred or not. This internal counter and
   the TSR are both reset whenever a new value is written to the TIR.

   This internal counter gets incremented on every tick of
   `SYSTEM_CLOCK', so the TIR value has to be multiplied by some
   constant factor (SYSTEM_CLOCK freq / 1 KHz) so that the timer goes
   off after the correct number of milliseconds. This allows this
   module to use just one clock, which makes things a lot simpler, but
   means you have to update this module manually whenever the clock
   frequency changes.
 
 */


module timer_device ( selectStatus,
                      selectInterval,
                      in, // value to write to interval register
                      out, // output of interval or status register (depending on `select*')
                      we, // whether the interval is being set
                      gwe, // global we
                      reset, // global reset

                      // I/O pin connections
                      SYSTEM_CLOCK);

   input [15:0]   in;
   input          we, gwe;
   input          selectStatus, selectInterval;
   input          reset;
   input          SYSTEM_CLOCK;
   output [15:0]  out;
   
   wire [15:0]    in, out;
   wire           we, gwe, reset, SYSTEM_CLOCK;
   wire           selectStatus, selectInterval;

   wire [15:0]    interval_out;
   wire           status_in, status_out;
   wire [31:0]    counter_in, counter_out, incremented_counter;
   wire           timerHasGoneOff;
   
   wire           intervalBeingWritten;
   and( intervalBeingWritten, selectInterval, we );

   // INTERVAL REGISTER
   
   Nbit_reg #(16, 0) interval (.in( in ),
                                   .out( interval_out ),
                                   .clk( SYSTEM_CLOCK ),
                                   .we( intervalBeingWritten ),
                                   .gwe( gwe ),
                                   .rst( reset ));

   // COUNTER
   
   mux_Nbit_2to1 #(32) counterInputMux (.a( incremented_counter ),
                                        .b( 32'h0000 ),
                                        .sel( timerHasGoneOff ), // timer rolls itself over
                                        .out( counter_in ));
   
   Nbit_reg #(32, 0) counter (.in( counter_in ),
                                  .out( counter_out ),
                                  .clk( SYSTEM_CLOCK ),
                                  .we( 1'b1 ),
                                  .gwe( gwe ),
                                  .rst( reset | intervalBeingWritten ));

   Nbit_incrementer #(32) incr_inst (.in( counter_out ),
                                     .out( incremented_counter ));

   wire [31:0]    timeUntilNextTick;
   /* UPDATE SYSTEM_CLOCK FREQUENCY HERE: we "multiply" the interval to 
      translate the interval into processor ticks */ 
   // 8 MHz clock -> 8Mega ticks per ms -> 2^13 ticks per ms
   // log_2(2^13) -> 13
   // Since interval_out is 16 bits, pad the shifted value out to 32 bits: 32 - 16 - 13 -> 3
   Nbit_subtracter #(32) sub_inst (.a( { {3{1'b0}}, interval_out, {13{1'b0}} } ),
                                   .b( counter_out ),
                                   .difference( timeUntilNextTick )); // a - b
   // timeUntilNextTick <= 0 means we should have a tick!
   or( timerHasGoneOff, &(~timeUntilNextTick), timeUntilNextTick[31] ); 

   
   // STATUS REGISTER
   
   mux_Nbit_4to1 #(1) statusInputMux (.a( status_out ), // maintain old status value
                                      .b( 1'b0 ), // status is being read, so reset it
                                      .c( 1'b1 ), // timer went off
                                      .d( 1'b0 ), /* timer went off AND status is
                                                     being read, don't want to lose 
                                                     the "tick" */
                                      .sel( {timerHasGoneOff, selectStatus} ),
                                      .out( status_in ));

   Nbit_reg #(1, 0) status (.in( status_in ),
				 .out( status_out ),
				 .clk( SYSTEM_CLOCK ),
				 .we( 1'b1 ),
				 .gwe( gwe ),
				 .rst( reset | intervalBeingWritten ));

   mux_Nbit_4to1 #(16) outputMux(.a( 16'h0000 ), // sel == 00 - NOOP
                                 .b( { status_out, {15{1'b0}} } ), // sel == 01
                                 .c( interval_out ), // sel == 10
                                 .d( 16'h0000 ), // sel = 11 - INVALID
                                 .sel( {selectInterval, selectStatus} ),
				 .out( out ));

endmodule